var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

//Export types
export { default as union } from './types/union';
export { default as tuple } from './types/tuple';
export { default as Any } from './types/any';
export { default as Nil } from './types/nil';
export { default as reference } from './types/reference';
export { default as ObjectId } from './types/object-id';
export { default as collection } from './types/collection';
export { default as collections } from './types/collections';
export { default as model, makeOwner } from './types/model';

//Export modifiers
export { default as optional } from './modifiers/optional';
export { default as validate } from './modifiers/validate';
export { default as bare } from './modifiers/bare';
export { default as reducer } from './modifiers/reducer';
export { default as wrapGenerator } from './modifiers/wrap-generator';
export { default as autoResolve } from './modifiers/auto-resolve';

//Export generic type parser
export { default as type } from './parse/type';

import Store from './store';

export default function reduxSchemaStore(schema, options, createStore, preloadedState, enhancer) {
  var result = function result(createStore) {
    return function (reducer, preloadedState, enhancer) {
      var store = new Store(_extends({ schema: schema }, options)),
          redux = void 0;

      if (preloadedState !== undefined) {
        var message = store.schema.validateData(preloadedState);
        if (message) throw new TypeError('Can\'t use preloaded state: ' + message);
      }

      if (reducer) {
        (function () {
          var customReducer = reducer;
          reducer = function reducer(state, action) {
            return customReducer(store.reducer(state, action), action);
          };
        })();
      } else {
        reducer = store.reducer;
      }

      redux = createStore(reducer, preloadedState, enhancer);

      store.getState = redux.getState;
      store.replaceReducer = redux.replaceReducer;
      store.subscribe = redux.subscribe;

      store.store = redux;

      return store;
    };
  };

  return createStore ? result(createStore)(null, preloadedState, enhancer) : result;
}