var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

import { snakeCase, pathToStr, namedFunction, toObject } from '../utils';
import co from 'co';
import { isGeneratorFunction } from '../utils';

function freezeObject(obj) {
  if (Object.freeze) {
    Object.freeze(obj);
  }
}

export function hydratePrototype(_ref) {
  var type = _ref.type,
      typePath = _ref.typePath,
      getter = _ref.getter,
      setter = _ref.setter,
      keys = _ref.keys,
      _ref$properties = _ref.properties,
      properties = _ref$properties === undefined ? {} : _ref$properties,
      _ref$methods = _ref.methods,
      methods = _ref$methods === undefined ? {} : _ref$methods,
      _ref$virtuals = _ref.virtuals,
      virtuals = _ref$virtuals === undefined ? {} : _ref$virtuals,
      _ref$meta = _ref.meta,
      meta = _ref$meta === undefined ? {} : _ref$meta,
      freeze = _ref.freeze,
      namedFunctions = _ref.namedFunctions,
      _ref$wrapGenerators = _ref.wrapGenerators,
      wrapGenerators = _ref$wrapGenerators === undefined ? false : _ref$wrapGenerators,
      generatorWrapper = _ref.generatorWrapper;

  var prototype = Object.create(type.kind === 'array' ? Array.prototype : Object.prototype),
      define = {},
      typeSnake = snakeCase(pathToStr(typePath)).replace('.', '_');

  meta = _extends({}, meta, {
    type: type,
    typePath: typePath,
    recordRead: function recordRead() {
      this.store.recordRead(this.storePath);
    },

    scratch: {}
  });

  Object.defineProperties(meta, {
    state: {
      get: function get() {
        return this.store.get(this.storePath);
      },

      set: function set(value) {
        this.store.put(this.storePath, value);
      }
    },
    options: {
      get: function get() {
        return this.type && this.type.options;
      }
    }
  });

  define.get = {
    enumerable: true,
    value: function value(propName) {
      if (virtuals[propName]) return this[propName];
      var meta = this._meta;
      return getter.call(this, propName, meta.store, meta.storePath, meta.instancePath, this);
    }
  };

  define.set = {
    enumerable: true,
    value: function value(propName, _value) {
      if (virtuals[propName]) {
        this[propName] = _value;
        return;
      }
      var meta = this._meta;
      return setter.call(this, propName, _value, meta.store, meta.storePath, meta.instancePath, this);
    }
  };

  define.keys = {
    enumerable: false,
    get: keys
  };

  if (freeze) {
    freezeObject(meta);
  }

  define._meta = {
    enumerable: false,
    value: meta
  };

  define.toObject = {
    enumerable: false,
    value: function value() {
      return toObject(this);
    }
  };

  define.inspect = {
    enumerable: false,
    value: function value() {
      return toObject(this);
    }
  };

  Object.keys(properties).forEach(function (propName) {
    define[propName] = {
      enumerable: true,
      get: function get() {
        var meta = this._meta;
        return getter.call(this, propName, meta.store, meta.storePath, meta.instancePath, meta.instance);
      },
      set: function set(value) {
        var meta = this._meta;
        return setter.call(this, propName, value, meta.store, meta.storePath, meta.instancePath, meta.instance);
      }
    };
  });

  Object.keys(methods).forEach(function (methodName) {
    var invokeName = methodName,
        method = methods[methodName],
        actionType = (typeSnake ? typeSnake + '_' : '') + snakeCase(methodName),
        func = method,
        wrap = void 0,
        wrapper = void 0;

    wrapper = generatorWrapper || co.wrap;
    wrap = 'wrapGenerator' in method ? method.wrapGenerator : wrapGenerators;

    if (typeof wrap === 'function') {
      wrapper = wrap;
    }

    if (wrap && isGeneratorFunction(func)) {
      func = wrapper(func);
    }

    if (method.noWrap) {
      define[methodName] = {
        enumerable: true,
        value: func
      };
    } else {
      if (methodName === 'constructor') {
        invokeName = typePath[0];
      }
      define[methodName] = {
        enumerable: true,
        value: namedFunction(invokeName, function invokeMethod() {
          var meta = this._meta,
              path = meta.instancePath.concat(methodName);

          for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
            args[_key] = arguments[_key];
          }

          return meta.store.invoke(this, actionType, path, func, args);
        }, method, !namedFunctions)
      };
    }
  });

  Object.keys(virtuals).forEach(function (virtualName) {
    var actionType = 'SET_' + typeSnake + '_' + snakeCase(virtualName),
        prop = virtuals[virtualName];

    define[virtualName] = {
      enumerable: true,
      get: prop.get,
      set: function set(value) {
        var meta = this._meta,
            propInstancePath = meta.instancePath.concat(virtualName);
        meta.store.setVirtual(this, actionType, propInstancePath, prop.set, value);
      }
    };
  });

  Object.defineProperties(prototype, define);
  if (freeze) {
    freezeObject(prototype);
  }
  return prototype;
}

export function hydrateInstance(_ref2) {
  var prototype = _ref2.prototype,
      store = _ref2.store,
      storePath = _ref2.storePath,
      instancePath = _ref2.instancePath,
      currentInstance = _ref2.currentInstance,
      meta = _ref2.meta,
      freeze = _ref2.freeze;

  var instance = currentInstance || Object.create(prototype);

  meta = _extends(Object.create(instance._meta), meta);

  meta.store = store;
  meta.storePath = storePath;
  meta.instancePath = instancePath;

  if (freeze) {
    freezeObject(meta);
  }

  Object.defineProperty(instance, '_meta', {
    enumerable: false,
    value: meta
  });

  if (freeze) {
    freezeObject(instance);
  }

  return instance;
}