var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

import { snakeCase, pathToStr } from './utils';
import deepMerge from 'deepmerge';

var freeze = Object.freeze ? Object.freeze.bind(Object) : function () {};

var Store = function () {
  function Store(options) {
    _classCallCheck(this, Store);

    function error() {
      throw new Error('Schema Store has no Redux Store assigned');
    }

    var _typeMoniker$skipWrit = _extends({
      typeMoniker: [],
      skipWriteSame: false
    }, options, {
      store: this
    }),
        schema = _typeMoniker$skipWrit.schema,
        newOptions = _objectWithoutProperties(_typeMoniker$skipWrit, ['schema']);

    options = newOptions;

    if (options.debug) {
      options.freeze = true;
      options.validate = true;
      options.namedFunctions = true;
    }

    if (!schema || !schema.isType) {
      throw new Error('Missing schema in Store options');
    }

    this.result = undefined;
    this.internalState = undefined;
    this.reducer = this.reducer.bind(this);
    this.store = {
      dispatch: error,
      getState: error
    };
    this.options = options;
    this.schema = schema(options);
    this.maxCache = 1024;
    this.cache = {};
    this.cachePaths = [];
    this.record = null;
    this.recordStack = [];
    this.traceSuspended = 0;
  }

  Store.prototype.setVirtual = function setVirtual(obj, actionType, instancePath, setter, value) {
    var action = {
      type: actionType,
      path: instancePath,
      arg: value
    };

    if (this.internalState === undefined) {
      this.store.dispatch(action);
      var result = this.result;
      this.result = undefined;
      return result;
    } else {
      if (this.verifyAction && this.verifyAction.type !== actionType) return;
      this.verifyAction = null;
      return setter.apply(obj, [value]);
    }
  };

  Store.prototype.invoke = function invoke(obj, actionType, instancePath, method, args) {
    var self = this,
        action = {
      type: actionType,
      path: instancePath,
      args: args
    },
        currentState = void 0,
        newState = void 0,
        result = void 0,
        arg = void 0,
        ix = void 0;

    function argResolved(value) {
      args = args.slice();
      args[ix] = value;
      return self.invoke(obj, actionType, instancePath, method, args);
    }

    if (method.autoResolve) {
      for (ix = 0; ix < args.length; ix++) {
        arg = args[ix];
        if (arg && typeof arg === 'object' && typeof arg.then === 'function') {
          return arg.then(argResolved);
        }
      }
    }

    if (this.internalState === undefined) {
      this.store.dispatch(action);
      result = this.result;
      this.result = undefined;
      return result;
    } else {
      if (this.verifyAction && this.verifyAction.type !== actionType) return;
      this.verifyAction = null;
      if (method.reducer) {
        result = {};

        currentState = this.get(obj._meta.storePath);

        newState = method.call(obj, currentState, args, result);

        if (currentState !== newState) {
          this.put(obj._meta.storePath, newState);
        }
        if ('result' in result) return result.result;
        return obj;
      }
      return method.apply(obj, args);
    }
  };

  Store.prototype.put = function put(path, value) {
    this.checkRecord();

    if (this.options.skipWriteSame) {
      if (this.get(path) === value) return;
    }

    var action = {
      type: 'SET_' + (this.propActionFromPath(path) || 'ROOT'),
      path: path,
      value: value
    };

    if (this.internalState === undefined) {
      this.store.dispatch(action);
      var result = this.result;
      this.result = undefined;
      return result;
    } else {
      return this.executeAction(action);
    }
  };

  Store.prototype.get = function get(path) {
    this.recordRead(path);
    var toGo = path.slice(),
        current = this.getActiveState();
    while (current && toGo.length) {
      current = current[toGo.shift()];
    }
    return current;
  };

  Store.prototype.reducer = function reducer(state, action) {
    if (state === undefined) {
      state = this.schema.defaultValue();
    }
    if (!action.path) return state;
    this.internalState = state;
    try {
      this.result = this.executeAction(action);
      state = this.internalState;
      if (this.options.freeze) {
        freeze(state);
      }
    } finally {
      this.internalState = undefined;
    }
    return state;
  };

  Store.prototype.getActiveState = function getActiveState() {
    if (this.internalState !== undefined) {
      return this.internalState;
    }
    return this.store.getState();
  };

  Store.prototype.executeAction = function executeAction(action) {

    function updateProperty(state, path, value) {
      if (!path.length) return value;
      var name = path[0],
          prop = state[name],
          updated = updateProperty(prop, path.slice(1), value),
          newState = void 0;

      if (updated === prop) return state;

      if (Array.isArray(state) && /^\-?\d+$/.test(name)) {
        newState = state.slice();
        newState[Number(name)] = updated;
        return newState;
      } else {
        if (Array.isArray(state)) {
          if (name === 'length') {
            newState = state.slice();
            newState.length = updated;
            return newState;
          }
          throw new Error('Property put does not support extra properties on Arrays');
        }
        newState = _extends({}, state);
        if (updated === undefined) {
          delete newState[name];
        } else {
          newState[name] = updated;
        }
        return newState;
      }
    }

    var path = action.path.slice(),
        methodOrPropName = void 0,
        instance = void 0,
        result = void 0;

    if ('value' in action && action.type === 'SET_' + (this.propActionFromPath(path) || 'ROOT')) {
      this.checkRecord();
      this.internalState = updateProperty(this.internalState, path, action.value);
    } else {
      methodOrPropName = path.pop();
      try {
        instance = this.traversePath(path);
      } catch (err) {
        if (/_CONSTRUCTOR$/.test(action.type) && /not found in state/.test(err.message) && action.args) {
          this.put(action.path.slice(0, -1), {});
          instance = this.traversePath(path);
        } else {
          throw err;
        }
      }

      this.verifyAction = action;
      if (action.args) {
        result = instance[methodOrPropName].apply(instance, action.args);
      } else {
        result = instance[methodOrPropName] = action.arg;
      }
      this.verifyAction = null;
    }
    return result;
  };

  Store.prototype.traversePath = function traversePath(path) {
    var toGo = path.slice(),
        current = this.instance;
    while (current && toGo.length) {
      current = current.get(toGo.shift());
    }
    if (!current) {
      throw new Error('Path "' + path.join('.') + '" not found in state.');
    }
    return current;
  };

  Store.prototype.unpack = function unpack(type, storePath, instancePath, currentInstance, owner) {
    var path = void 0,
        cached = void 0,
        result = void 0,
        ix = void 0;
    result = type.unpack(this, storePath, instancePath, currentInstance, owner);
    if (!result || typeof result !== 'object') return result;

    path = pathToStr(instancePath);
    cached = this.cache[path];
    if (cached && !currentInstance && result._meta && result._meta.options && result._meta.options.typeMoniker.join('.') === cached._meta.options.typeMoniker.join('.')) {
      ix = this.cachePaths.indexOf(path);
      if (ix !== -1) {
        this.cachePaths.splice(ix, 1);
      }
      this.cachePaths.push(path);
      return cached;
    }

    this.cache[path] = result;
    this.cachePaths.push(path);
    if (this.cachePaths.length > this.maxCache) {
      delete this.cache[this.cachePaths.shift()];
    }
    return result;
  };

  Store.prototype.recordRead = function recordRead(path) {
    if (!this.record) return;
    var reg = this.record;

    for (var i = 0; i < path.length; i++) {
      var prop = path[i];

      if (!reg.children[prop]) {
        reg.children[prop] = { children: {} };
      }
      reg = reg.children[prop];
    }

    reg.check = true;
  };

  Store.prototype.startRecord = function startRecord() {
    var _this = this;

    if (this.record && !this.traceSuspended) {
      this.recordStack.push(this.record);
    }
    this.record = { children: {} };
    var current = this.record;
    if (this.options.debug) {
      var traceError = new Error('startRecord called without a matching stopRecord');
      process.nextTick(function () {
        if (_this.record === current) {
          throw traceError;
        }
      });
    }
  };

  Store.prototype.stopRecord = function stopRecord(remove) {
    var result = this.record,
        parent = this.recordStack.pop();

    if (!remove && parent) {
      deepMerge(parent, result);
    }

    if (parent) {
      this.record = parent;
    } else {
      this.record = null;
    }

    function readState(record, state) {
      var result = {
        state: state,
        children: {},
        check: record.check
      };

      Object.keys(record.children).forEach(function (prop) {
        return result.children[prop] = readState(record.children[prop], state[prop]);
      });

      return result;
    }

    return readState(result, this.state);
  };

  Store.prototype.sameRecordedState = function sameRecordedState(snapshot) {
    function compareState(snapshot, snapShotState, state) {
      if (state === snapShotState) return true;
      if (snapshot.check) {
        if (!state || !snapShotState || typeof state !== 'object' || typeof snapShotState !== 'object') return false;
        //Only compare this object itself, which only includes the names of the properties. Their values have their own trackers.
        for (var prop in state) {
          if (!(prop in snapShotState)) return false;
        }for (var _prop in snapShotState) {
          if (!(_prop in state)) return false;
        }
      }
      return Object.keys(snapshot.children).every(function (prop) {
        return snapShotState && state && compareState(snapshot.children[prop], snapShotState[prop], state[prop]);
      });
    }

    return compareState(snapshot, snapshot.state, this.state);
  };

  Store.prototype.trace = function trace(func, remove) {
    var result = void 0;
    this.startRecord();
    try {
      func();
    } finally {
      result = this.stopRecord(remove);
    }
    return result;
  };

  Store.prototype.suspendTrace = function suspendTrace(func) {
    this.traceSuspended++;
    try {
      func();
    } finally {
      this.traceSuspended--;
    }
  };

  Store.prototype.checkRecord = function checkRecord() {
    if (this.record) {
      throw new Error('Cannot write state while in read recording mode');
    }
  };

  Store.prototype.propActionFromPath = function propActionFromPath(path) {
    return snakeCase(pathToStr(this.schema.getTypeFromPath(path)).replace(/\./g, '_'));
  };

  _createClass(Store, [{
    key: 'instance',
    get: function get() {
      if (!this._instance) {
        this._instance = this.unpack(this.schema, [], []);
      }
      return this._instance;
    },
    set: function set(value) {
      this.put([], this.schema.pack(value));
    }
  }, {
    key: 'state',
    get: function get() {
      return this.getActiveState();
    },
    set: function set(value) {
      var message = this.schema.validateData(value);
      if (message) throw new TypeError('Can\'t assign state: ' + message);
      this.put([], value);
    }
  }, {
    key: 'dispatch',
    get: function get() {
      return this.store.dispatch;
    },
    set: function set(value) {
      this.store.dispatch = value;
    }
  }, {
    key: 'models',
    get: function get() {
      return this.instance.models;
    }
  }, {
    key: 'isStore',
    get: function get() {
      return true;
    }
  }]);

  return Store;
}();

export default Store;