import finalizeType from '../parse/finalize-type';
import { pathToStr } from '../utils';

export default function reference(target) {
  function Reference(options) {
    return finalizeType({
      isType: true,
      name: pathToStr(options.typeMoniker),
      kind: 'reference',
      storageKinds: ['string'],
      options: options,
      validateData: function validateData(value, instancePath) {
        instancePath = instancePath || options.typeMoniker;
        if (typeof value !== 'string') {
          return 'Reference data for "' + pathToStr(instancePath) + '" must be a string';
        }
        if (!value) return 'Reference cannot be empty';
      },
      validateAssign: function validateAssign(value, instancePath) {
        instancePath = instancePath || options.typeMoniker;
        if (typeof value !== 'object' || !value._meta || !value._meta.idKey) {
          return 'Reference for "' + pathToStr(instancePath) + '" must be an object of type "' + target + '"';
        }
      },
      pack: function pack(value) {
        return value[value._meta.idKey];
      },
      unpack: function unpack(store, storePath, instancePath, currentInstance, owner) {
        var findCollection = function findCollection() {
          var ancestor = owner,
              found = void 0,
              type = void 0,
              collection = void 0;

          while (true) {
            if (!ancestor || !ancestor._meta) break;
            type = ancestor._meta.type;
            if (!type) break;
            if (type.isCollections && type.properties[target] && type.properties[target].isCollection) {
              collection = ancestor[target];
              if (collection && collection._meta && (type = collection._meta.type) && type.model) {
                found = collection;
                break;
              }
            }
            ancestor = ancestor._meta.owner;
          }
          return found;
        };

        var id = store.get(storePath);
        if (!id || id === '<unknown>') throw new ReferenceError('Cannot dereference: No "' + target + '" id present');

        var collection = findCollection();

        if (!collection) throw new TypeError('Cannot find collection of type "' + target + '"');

        var result = collection.get(id);
        if (!result) throw new ReferenceError('Cannot dereference: No "' + target + '" with id "' + id + '" exists');
        return result;
      },
      getTypeFromPath: function getTypeFromPath(path) {
        if (path.length) throw new Error('Type paths for references are not supported');
        return options.typeMoniker;
      },
      defaultValue: function defaultValue() {
        return '<unknown>';
      }
    });
  }

  Reference.isType = true;
  return Reference;
}