'use strict';

exports.__esModule = true;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports['default'] = parseObjectType;

var _any = require('../types/any');

var _any2 = _interopRequireDefault(_any);

var _finalizeType = require('../parse/finalize-type');

var _finalizeType2 = _interopRequireDefault(_finalizeType);

var _parseType = require('../parse/parse-type');

var _parseType2 = _interopRequireDefault(_parseType);

var _utils = require('../utils');

var _basic = require('../types/basic');

var _array = require('./array');

var _hydrate = require('./hydrate');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function parseObjectType(options, type, arrayType) {
  if (type === Object) return (0, _parseType2['default'])(options, {});
  if (type === Array) return (0, _parseType2['default'])(options, []);

  if (typeof type !== 'object') throw new TypeError((0, _utils.pathToStr)(options.typeMoniker) + ' type must be an object');

  arrayType = Boolean(arrayType);

  var self = { options: options },
      typeMoniker = options.typeMoniker,
      propNames = Object.keys(type),
      properties = {},
      virtuals = {},
      methods = {},
      meta = {},
      kind = arrayType ? type.length === 1 ? 'array' : 'tuple' : 'object',
      storedKeys = [],
      storedState = void 0,
      prototype = void 0,
      thisType = void 0,
      restType = void 0;

  if (options.self) {
    self = _extends(options.self, self);
  }

  if (arrayType) {
    if (!type.length) {
      return (0, _parseType2['default'])(options, Array);
    }

    if (type.length === 1) {
      restType = (0, _parseType2['default'])(_extends({}, options, { parent: self, self: null, typeMoniker: typeMoniker.concat('*') }), type[0]);
      propNames = [];
      methods = _extends({}, _array.arrayMethods);
      virtuals = _extends({}, _array.arrayVirtuals);
    }
  } else {
    if (!propNames.length) {
      return (0, _parseType2['default'])(options, Object);
    }

    if (propNames.indexOf('*') !== -1) {
      propNames.splice(propNames.indexOf('*'), 1);
      if (!propNames.length && type['*'] === _any2['default']) return (0, _parseType2['default'])(options, Object);
      restType = (0, _parseType2['default'])(_extends({}, options, { parent: self, self: null, typeMoniker: typeMoniker.concat('*') }), type['*']);
    }
  }

  propNames.forEach(function (prop) {
    var descriptor = Object.getOwnPropertyDescriptor(type, prop);
    if (descriptor.get || descriptor.set || (typeof descriptor.value === 'function' || descriptor.value && typeof descriptor.value.method === 'function') && !(0, _basic.functionIsType)(descriptor.value)) {
      if (descriptor.value) {
        methods[prop] = descriptor.value;
      } else {
        var virtual = {};
        if (descriptor.get) {
          virtual.get = descriptor.get;
        }
        if (descriptor.set) {
          virtual.set = descriptor.set;
        }
        virtuals[prop] = virtual;
      }
    } else {
      properties[prop] = (0, _parseType2['default'])(_extends({}, options, {
        parent: self,
        self: null,
        typeMoniker: typeMoniker.concat(prop)
      }), type[prop]);
      if (properties[prop].name === 'objectid' && !meta.idKey) {
        meta.idKey = prop;
      }
    }
  });

  propNames = Object.keys(properties);

  thisType = {
    isType: true,
    name: (0, _utils.pathToStr)(typeMoniker) || arrayType ? 'array' : 'object',
    kind: kind,
    storageKinds: [kind],
    options: options,
    validateData: function validateData(value, instancePath) {
      instancePath = instancePath || typeMoniker;
      if (typeof value !== 'object') {
        return 'Type of "' + (0, _utils.pathToStr)(instancePath) + '" data must be object';
      }
      return propNames.reduce(function (message, name) {
        return message || properties[name].validateData(value[name], instancePath.concat(name));
      }, null) || Object.keys(value).reduce(function (message, name) {
        if (message) return message;
        if (restType) {
          if (propNames.indexOf(name) !== -1) return null;
          return restType.validateData(value[name], instancePath.concat(name));
        } else {
          if (propNames.indexOf(name) === -1) {
            return 'Unknown data property "' + (0, _utils.pathToStr)(instancePath.concat(name)) + '"';
          }
        }
      }, null);
    },
    validateAssign: function validateAssign(value, instancePath) {
      instancePath = instancePath || typeMoniker;

      if (typeof value !== 'object') {
        return 'Type of "' + (0, _utils.pathToStr)(instancePath) + '" must be object';
      }

      var isSchemaObject = value && value._meta && value._meta.type && (value._meta.type.kind === 'object' || value._meta.type.kind === 'array');

      function getProp(prop) {
        if (isSchemaObject) {
          return value.get(prop);
        }
        return value[prop];
      }

      function getKeys(value) {
        if (isSchemaObject) {
          return value.keys;
        }
        return Object.keys(value);
      }

      instancePath = instancePath || typeMoniker;
      return propNames.reduce(function (message, name) {
        return message || properties[name].validateAssign(getProp(name), instancePath.concat(name));
      }, null) || getKeys(value).reduce(function (message, name) {
        if (message) return message;
        if (propNames.indexOf(name) !== -1) return null;
        if (restType) {
          return restType.validateAssign(getProp(name), instancePath.concat(name));
        } else {
          return 'Unknown property "' + (0, _utils.pathToStr)(instancePath.concat(name)) + '"';
        }
      }, null);
    },
    pack: function pack(value) {
      var isSchemaObject = value && value._meta && value._meta.type && (value._meta.type.kind === 'object' || value._meta.type.kind === 'array');

      function getProp(prop) {
        if (isSchemaObject) {
          return value.get(prop);
        }
        return value[prop];
      }

      function getKeys(value) {
        if (isSchemaObject) {
          return value.keys;
        }
        return Object.keys(value);
      }

      var out = arrayType ? [] : {};
      propNames.forEach(function (name) {
        return out[name] = properties[name].pack(getProp(name));
      });

      if (restType) {
        getKeys(value).forEach(function (name) {
          if (propNames.indexOf(name) === -1) {
            if (arrayType) {
              if (isNaN(name) || Number(name) % 1 !== 0 || Number(name) < 0 || String(Number(name)) !== String(name)) {
                throw new TypeError('Cannot set "' + (0, _utils.pathToStr)(options.typeMoniker.concat(name)) + '" property on array');
              }
            }
            out[name] = restType.pack(getProp(name));
          }
        });
      }
      return out;
    },
    unpack: function unpack(store, storePath, instancePath, currentInstance, owner) {
      return (0, _hydrate.hydrateInstance)(_extends({}, options, {
        prototype: prototype,
        store: store,
        storePath: storePath,
        instancePath: instancePath,
        currentInstance: currentInstance,
        meta: { owner: owner }
      }));
    },
    getTypeFromPath: function getTypeFromPath(path) {
      if (!path.length) return typeMoniker;

      var first = path[0],
          type = void 0;

      if (propNames.indexOf(first) !== -1) {
        type = properties[first];
      } else {
        if (!restType) {
          if (this.virtuals[first] && path.length === 1) {
            return options.typeMoniker.concat(first);
          }
          throw new Error('Path not found: ' + (0, _utils.pathToStr)(typeMoniker.concat(path)));
        }
        type = restType;
      }
      return type.getTypeFromPath(path.slice(1));
    },
    defaultValue: function defaultValue() {
      var defaultValue = arrayType ? [] : {};
      Object.keys(properties).forEach(function (name) {
        return defaultValue[name] = properties[name].defaultValue();
      });
      return defaultValue;
    },

    properties: properties,
    restType: restType,
    methods: methods,
    virtuals: virtuals,
    getPropType: function getPropType(name) {
      if (propNames.indexOf(name) !== -1) {
        return properties[name];
      } else {
        if (!restType) throw new TypeError('Unknown property ' + (0, _utils.pathToStr)(typeMoniker.concat(name)));
        return restType;
      }
    },
    defaultRestProp: function defaultRestProp() {
      if (restType) return restType.defaultValue();
    },
    packProp: function packProp(name, value) {
      var type = void 0;
      if (propNames.indexOf(name) !== -1) {
        type = properties[name];
      } else {
        if (!restType) throw new TypeError('Unknown property ' + (0, _utils.pathToStr)(typeMoniker.concat(name)));
        type = restType;
      }
      return type.pack(value);
    }
  };

  if (arrayType) {
    thisType.length = type.length;
  }

  if ('name' in self) {
    delete thisType.name;
  }
  _extends(self, (0, _finalizeType2['default'])(thisType));

  prototype = (0, _hydrate.hydratePrototype)(_extends({}, options, {
    type: self,
    typePath: typeMoniker,
    getter: function getter(name) {
      var _this = this;

      var meta = this._meta,
          ix = Number(name),
          type = void 0;

      if (arrayType) {
        if (isNaN(name) || ix % 1 !== 0 || ix < 0 || String(ix) !== String(name) || ix >= this.length) {
          return undefined;
        }
      }

      if (propNames.indexOf(name) !== -1) {
        type = properties[name];
      } else {
        if (!restType) {
          throw new TypeError('Unknown property ' + (0, _utils.pathToStr)(typeMoniker.concat(name)));
        }
        type = restType;
        var hasKey = void 0;
        this._meta.store.suspendTrace(function () {
          return hasKey = _this.keys.indexOf(String(name)) !== -1;
        });
        if (!hasKey) return;
      }
      return meta.store.unpack(type, meta.storePath.concat(name), meta.instancePath.concat(name), null, this);
    },
    setter: function setter(name, value) {
      var meta = this._meta,
          ix = Number(name),
          newState = void 0,
          packed = void 0;

      name = String(name);

      if (propNames.indexOf(name) !== -1) {
        type = properties[name];
      } else {
        if (!restType) throw new TypeError('Unknown property ' + (0, _utils.pathToStr)(meta.instancePath.concat(name)));
        type = restType;
      }

      if (arrayType) {
        if (isNaN(name) || ix % 1 !== 0 || ix < 0 || String(ix) !== String(name)) {
          throw new TypeError('Cannot set "' + (0, _utils.pathToStr)(options.typeMoniker.concat(name)) + '" property on array');
        }
        newState = meta.store.get(meta.storePath);
        if (ix > this.length) {
          newState = newState.slice();
          while (ix > newState.length) {
            newState.push(type.defaultValue());
          }
          newState.push(type.pack(value));
          return meta.store.put(meta.storePath, newState);
        }
      }

      if (!arrayType && value === undefined && type === restType) {
        packed = undefined;
      } else {
        packed = type.pack(value);
      }

      meta.store.put(meta.storePath.concat(name), packed);
    },
    keys: function keys() {
      this._meta.store.recordRead(this._meta.storePath);
      var state = this._meta.state;
      if (storedState !== state) {
        storedKeys = Object.keys(state);
        storedState = state;
      }
      return storedKeys;
    },

    properties: properties,
    methods: methods,
    virtuals: virtuals,
    meta: meta
  }));

  self.prototype = prototype;

  return self;
}