'use strict';

exports.__esModule = true;
exports['default'] = parseType;

var _parseObject = require('./parse-object');

var _parseObject2 = _interopRequireDefault(_parseObject);

var _anyObject = require('./any-object');

var _anyObject2 = _interopRequireDefault(_anyObject);

var _regexp = require('../types/regexp');

var _regexp2 = _interopRequireDefault(_regexp);

var _date = require('../types/date');

var _date2 = _interopRequireDefault(_date);

var _error = require('../types/error');

var _error2 = _interopRequireDefault(_error);

var _union = require('../types/union');

var _union2 = _interopRequireDefault(_union);

var _basic = require('../types/basic');

var _utils = require('../utils');

var _validate = require('../modifiers/validate');

var _validate2 = _interopRequireDefault(_validate);

var _optional = require('../modifiers/optional');

var _optional2 = _interopRequireDefault(_optional);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function parseType(options, type) {
  if (typeof type === 'function' && type.isType && !type.storageKinds) return type(options);
  if (type && type.isType) return type;

  if (type === Object) return (0, _anyObject2['default'])(options, false);
  if (type === Array) return (0, _anyObject2['default'])(options, true);

  if (type === null) return _basic.basicTypes['null'](options);
  if (type === undefined) return _basic.basicTypes.undefined(options);

  if (type === Number) return _basic.basicTypes.number(options);
  if (type === Boolean) return _basic.basicTypes['boolean'](options); //eslint-disable-line dot-notation
  if (type === String) return _basic.basicTypes.string(options);

  if (type === RegExp) return (0, _regexp2['default'])(options);
  if (type === Date) return (0, _date2['default'])(options);
  if (type === Error) return (0, _error2['default'])(options);

  if (typeof type === 'object') {
    if ((0, _utils.isArray)(type)) {
      if (!type.length) {
        return (0, _anyObject2['default'])(options, true);
      } else if (type.length > 1) {
        return _union2['default'].apply(null, type)(options);
      } else {
        return (0, _parseObject2['default'])(options, type, true);
      }
    } else {

      if ((0, _basic.functionIsType)(type.type)) {
        try {
          var actual = parseType(options, type.type);
          if (type.validate) {
            actual = (0, _validate2['default'])(actual, type.validate)(options);
          }
          if (type.optional) {
            actual = (0, _optional2['default'])(actual)(options);
          }
          return actual;
        } catch (err) {
          //empty
        }
      }

      return (0, _parseObject2['default'])(options, type);
    }
  }

  throw new TypeError('Unknown type ' + type);
}