'use strict';

exports.__esModule = true;
exports['default'] = date;

var _finalizeType = require('../parse/finalize-type');

var _finalizeType2 = _interopRequireDefault(_finalizeType);

var _utils = require('../utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function date(options) {
  var name = (0, _utils.pathToStr)(options.typeMoniker) || 'date';
  return (0, _finalizeType2['default'])({
    isType: true,
    name: name,
    kind: 'date',
    storageKinds: ['string'],
    options: options,
    validateData: function validateData(value, instancePath) {
      instancePath = instancePath || options.typeMoniker;
      if (value !== '' && (typeof value !== 'string' || new Date(value).toJSON() !== value)) {
        return 'Type of "' + ((0, _utils.pathToStr)(instancePath) || name) + '" data must be Date string';
      }
    },
    validateAssign: function validateAssign(value, instancePath) {
      instancePath = instancePath || options.typeMoniker;
      if (!(value instanceof Date)) {
        return 'Type of "' + ((0, _utils.pathToStr)(instancePath) || name) + '" must be Date';
      }
    },
    pack: function pack(value) {
      return value.toJSON() || '';
    },
    unpack: function unpack(store, path, instancePath, currentInstance) {
      if (currentInstance) throw new Error('Date types cannot modify a data instance');
      return new Date(store.get(path) || 'Invalid Date');
    },
    getTypeFromPath: function getTypeFromPath(path) {
      if (path.length) throw new Error('Cannot get type path for properties of Dates');
      return options.typeMoniker;
    },
    defaultValue: function defaultValue() {
      return '';
    }
  });
}