'use strict';

exports.__esModule = true;
exports.makeOwner = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports['default'] = model;

var _utils = require('../utils');

var _parseType = require('../parse/parse-type');

var _parseType2 = _interopRequireDefault(_parseType);

var _objectId = require('./object-id');

var _objectId2 = _interopRequireDefault(_objectId);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

var isOwner = {};

function model(name, model) {
  var collection = name[0].toLowerCase() + name.substr(1);

  function Model() {
    var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : { typeMoniker: [] };

    if (typeof model !== 'object') throw new TypeError('model definitions must be objects');

    var rebuild = false,
        ResultModel = void 0,
        resultType = void 0,
        idKey = void 0;

    ResultModel = (0, _utils.namedFunction)(name, function () {
      for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      if (!(this instanceof ResultModel)) {
        return new (Function.prototype.bind.apply(ResultModel, [null].concat(args)))();
      }

      var owner = args[0],
          store = options.store,
          id = (0, _utils.guid)(24),
          storePath = void 0,
          instancePath = void 0;
      if (owner && owner.isOwner === isOwner && owner.owner && owner.owner._meta && owner.owner._meta.store) {
        args.shift();
        if (owner.id) {
          id = owner.id;
        }
        owner = owner.owner;
      } else {
        owner = null;
      }
      if (!store || !store.isStore) {
        throw new Error('Cannot create new ' + name + '. No store assigned');
      }
      if (!owner) {
        throw new Error('Cannot create new ' + name + '. New instances only allowed for models in a collection.');
      }

      storePath = owner._meta.storePath.concat(id);
      instancePath = owner._meta.instancePath.concat(id);
      store.unpack(resultType, storePath, instancePath, this, owner);
      this.constructor.apply(this, args);
    }, model.constructor, !options.namedFunctions);

    resultType = (0, _parseType2['default'])(_extends({}, options, { self: ResultModel }), model);

    Object.keys(resultType.properties || {}).forEach(function (name) {
      if (resultType.properties[name].name === 'objectid') {
        idKey = name;
      }
    });

    if (!resultType.methods.hasOwnProperty('constructor')) {
      model.constructor = function (values) {
        var _this = this;

        if (values) {
          Object.keys(values).forEach(function (prop) {
            return _this.set(prop, values[prop]);
          });
        }
      };
      rebuild = true;
    }

    var origConstructor = model.constructor;

    model.constructor = (0, _utils.namedFunction)(name, function () {
      var storeInstance = void 0,
          path = this._meta.storePath;

      storeInstance = resultType.defaultValue();
      storeInstance[idKey] = path[path.length - 1];
      this._meta.store.put(path, storeInstance);

      if (origConstructor) {
        origConstructor.apply(this, arguments);
      }
    }, origConstructor, !options.namedFunctions);

    if (!idKey) {
      if (resultType.properties.id) {
        throw new TypeError('The "id" property of "' + name + '" must be an ObjectId');
      }
      model.id = _objectId2['default'];
      idKey = 'id';
      rebuild = true;
    }
    if (rebuild) {
      resultType = (0, _parseType2['default'])(_extends({}, options, { self: ResultModel }), model);
    }

    _extends(ResultModel, resultType, {
      prototype: resultType.prototype,
      collection: collection,
      idKey: idKey,
      model: model,
      isModel: true
    });

    return ResultModel;
  }

  Model.collection = collection;
  Model.modelName = name;
  Model.isType = true;
  Model.isModel = true;
  return Model;
}

var makeOwner = exports.makeOwner = function makeOwner(owner, id) {
  return {
    isOwner: isOwner,
    owner: owner,
    id: id
  };
};