'use strict';

exports.__esModule = true;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports['default'] = union;

var _finalizeType = require('../parse/finalize-type');

var _finalizeType2 = _interopRequireDefault(_finalizeType);

var _parseType = require('../parse/parse-type');

var _parseType2 = _interopRequireDefault(_parseType);

var _utils = require('../utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function union() {
  for (var _len = arguments.length, types = Array(_len), _key = 0; _key < _len; _key++) {
    types[_key] = arguments[_key];
  }

  if (!types.length) throw new TypeError('Union requires subtypes.');

  function Union(options) {
    var typeMoniker = options.typeMoniker,
        simple = true,
        storageKinds = {},
        kinds = {},
        handlersById = {},
        self = { options: options },
        handlerIds = void 0,
        handlers = void 0;

    function parse(type, name) {
      var moniker = [].concat(options.typeMoniker);
      if (name) {
        moniker.push(name);
      }
      return (0, _parseType2['default'])(_extends({}, options, {
        typeMoniker: moniker,
        parent: options.self || self, self: null
      }), type);
    }

    //Flatten all unions: union(union(null, undefined), union(Number, String)) => union(null, undefined, Number, String)
    types = Array.prototype.concat.apply([], types.map(function (type) {
      var handler = parse(type);
      if (handler.kind !== 'union') {
        return type;
      }
      return handler.types;
    }));

    handlers = types.map(function (type) {
      return parse(type);
    }).map(function (handler, ix) {
      handler.storageKinds.forEach(function (kind) {
        if (storageKinds[kind]) {
          simple = false;
        } else {
          storageKinds[kind] = true;
        }
      });
      if (!kinds[handler.kind]) {
        kinds[handler.kind] = 0;
      }
      kinds[handler.kind]++;
      var id = handler.kind + kinds[handler.kind];
      handler = parse(types[ix], id);
      handlersById[id] = handler;
      return handler;
    });

    handlerIds = Object.keys(handlersById);

    function idOfHandler(handler) {
      for (var i = 0; i < handlerIds.length; i++) {
        if (handlersById[handlerIds[i]] === handler) return handlerIds[i];
      }
    }

    var thisType = {
      isType: true,
      name: 'union(' + handlers.map(function (handler) {
        return handler.kind;
      }).join(', ') + ')',
      kind: 'union',
      storageKinds: Object.keys(storageKinds),
      options: options,
      validateData: function validateData(value, instancePath) {
        instancePath = instancePath || typeMoniker;
        if (!simple) {
          var keys = Object.keys(value || {}),
              type = keys[0];
          if (!value || keys.length !== 1) {
            return 'Missing union type for union "' + (0, _utils.pathToStr)(instancePath) + '"';
          }
          if (!handlersById[type]) {
            return 'Unexpected type "' + type + '" for union "' + (0, _utils.pathToStr)(instancePath) + '"';
          }
          return handlersById[type].validateData(value[type], instancePath);
        }
        return handlers.every(function (handler) {
          return handler.validateData(value, instancePath);
        }) && 'No matching data type for union "' + (0, _utils.pathToStr)(instancePath) + '"';
      },
      validateAssign: function validateAssign(value, instancePath) {
        instancePath = instancePath || typeMoniker;
        var messages = [],
            handler = void 0,
            message = void 0;

        for (var i = 0; i < handlers.length; i++) {
          message = handlers[i].validateAssign(value);
          if (!message) {
            handler = handlers[i];
            break;
          }
          messages.push(message);
        }
        if (!handler) {
          return 'Incompatible value ' + value + ' for ' + (0, _utils.pathToStr)(instancePath) + '. ' + messages.join(' -or- ') + '.';
        }
      },
      pack: function pack(value) {
        var _ref;

        var messages = [],
            handler = void 0,
            packed = void 0,
            message = void 0;
        for (var i = 0; i < handlers.length; i++) {
          message = handlers[i].validateAssign(value);
          if (!message) {
            handler = handlers[i];
            break;
          }
          messages.push(message);
        }
        if (!handler) {
          throw new TypeError('Incompatible value ' + value + ' for ' + thisType.name + '. ' + messages.join(' -or- ') + '.');
        }
        packed = handler.pack(value);
        if (simple) {
          return packed;
        }
        return _ref = {}, _ref[idOfHandler(handler)] = packed, _ref;
      },
      unpack: function unpack(store, path, instancePath, currentInstance, owner) {
        if (currentInstance) throw new Error('Union types cannot modify a data instance');
        var value = store.get(path);
        if (simple) {
          for (var i = 0; i < handlers.length; i++) {
            if (!handlers[i].validateData(value, instancePath)) {
              return store.unpack(handlers[i], path, instancePath, null, owner);
            }
          }
          throw new TypeError('No matching data type for union "' + (0, _utils.pathToStr)(instancePath) + '"');
        } else {
          var type = Object.keys(value || {})[0];
          if (!handlersById[type]) {
            if (type) {
              throw new Error('Unexpected type "' + type + '" for union "' + (0, _utils.pathToStr)(instancePath) + '"');
            } else {
              throw new Error('Missing union type for union "' + (0, _utils.pathToStr)(instancePath) + '"');
            }
          }
          return store.unpack(handlersById[type], path.concat(type), instancePath, null, owner);
        }
      },
      getTypeFromPath: function getTypeFromPath(path) {
        if (!path.length) {
          return options.typeMoniker;
        }
        var first = path[0];
        if (!handlersById[first]) throw new Error('Can\'t find union subtype ' + first + ', ' + path + ', ' + handlers[0].options.typeMoniker);
        return handlersById[first].getTypeFromPath(path.slice(1));
      },
      defaultValue: function defaultValue() {
        var _ref2;

        return simple ? handlers[0].defaultValue() : (_ref2 = {}, _ref2[idOfHandler(handlers[0])] = handlers[0].defaultValue(), _ref2);
      },

      handlers: handlers,
      types: types
    };

    return _extends(self, (0, _finalizeType2['default'])(thisType));
  }

  Union.isType = true;
  return Union;
}