'use strict';

exports.__esModule = true;
exports.snakeCase = undefined;
exports.isArray = isArray;
exports.isPlainObject = isPlainObject;
exports.guid = guid;
exports.pathToStr = pathToStr;
exports.namedFunction = namedFunction;
exports.toObject = toObject;
exports.dedent = dedent;
exports.isGenerator = isGenerator;
exports.isGeneratorFunction = isGeneratorFunction;

var _uuid = require('uuid');

var _uuid2 = _interopRequireDefault(_uuid);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function isArray(arr) {
  //Adapted from extend, Copyright (c) 2014 Stefan Thomas, MIT license
  if (typeof Array.isArray === 'function') {
    return Array.isArray(arr);
  }
  return Object.prototype.toString.call(arr) === '[object Array]';
}

function isPlainObject(obj) {
  //Adapted from extend, Copyright (c) 2014 Stefan Thomas, MIT license
  if (!obj || Object.prototype.toString.call(obj) !== '[object Object]') {
    return false;
  }

  var hasOwnConstructor = Object.prototype.hasOwnProperty.call(obj, 'constructor'),
      hasIsPrototypeOf = obj.constructor && obj.constructor.prototype && Object.prototype.hasOwnProperty.call(obj.constructor.prototype, 'isPrototypeOf'),
      key = void 0;

  if (obj.constructor && !hasOwnConstructor && !hasIsPrototypeOf) {
    return false;
  }
  for (key in obj) {/**/}
  return typeof key === 'undefined' || Object.prototype.hasOwnProperty.call(obj, key);
}

function guid(len) {
  var id = '';
  len = len || 24;

  while (id.length < len) {
    id += _uuid2['default'].v1().replace(/-/g, '');
  }
  return id.substr(0, len);
}

var snakeCase = exports.snakeCase = function snakeCase(camelCase) {
  return String(camelCase).replace(/[A-Z]/g, function (v) {
    return '_' + v;
  }).toUpperCase().replace(/^_/, '');
};

function pathToStr(path, delParams) {
  var result = '',
      item = void 0;
  for (var i = 0; i < path.length; i++) {
    item = path[i];
    if (delParams) {
      item = item.replace(/\(.*/, '');
    }
    if (i === 0 || /^[a-z$_][0-9a-z$_]*$/i.test(item)) {
      result += '.' + path[i];
    } else {
      if (String(Math.round(Number(item))) === item) {
        result += '[' + item + ']';
      } else {
        result += '[' + JSON.stringify(String(item)) + ']';
      }
    }
  }
  return result.replace(/^\./, '').replace(/\[\"\*\"\]/g, '[]');
}

function namedFunction(name, actualFunc, templateFunc, passThrough) {
  if (passThrough) {
    return actualFunc;
  }
  if (!templateFunc) {
    templateFunc = actualFunc;
  }

  var f = actualFunc // eslint-disable-line
  ,
      funcText = void 0,
      signature = void 0,
      func = void 0;

  if (typeof actualFunc !== 'function') throw new TypeError('Parameter to namedFunction must be a function');

  funcText = templateFunc.toString();

  signature = /\([^\)]*\)/.exec(funcText)[0];

  funcText = funcText.replace(/^[^{]+\{|}$/g, '');

  func = eval('(function(){function ' + (name + signature) + ' {/*' + funcText.replace(/\*\//g, '* /') + '*/return f.apply(this,arguments);}return ' + name + '}())'); // eslint-disable-line
  func.toString = function () {
    return 'function ' + (name + signature) + '{' + funcText + '}';
  };
  return func;
}

function toObject(value) {

  var seenValues = [],
      seenObjects = [];

  function internal(value) {
    if (value instanceof Date || value instanceof RegExp || value instanceof Error) return value;
    if (typeof value !== 'object' || !value) return value;
    if (!('keys' in value) || !('_meta' in value)) return value;
    if (value._meta.type.kind === 'array') return value.slice().map(function (v) {
      return internal(v);
    });
    var result = {},
        ix = seenValues.indexOf(value);

    if (ix !== -1) {
      return seenObjects[ix];
    }

    seenValues.push(value);
    seenObjects.push(result);

    value.keys.forEach(function (key) {
      result[key] = internal(value.get(key));
    });
    return result;
  }

  return internal(value);
}

function dedent(strings) {
  for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    args[_key - 1] = arguments[_key];
  }

  var string = strings.map(function (str, i) {
    return (i === 0 ? '' : args[i - 1]) + str;
  }).join(''),
      match = /^\n( *)/.exec(string),
      len = void 0,
      replace = void 0;

  if (!match) return string;

  len = match[1].length;

  replace = new RegExp('\\n {' + len + '}', 'g');

  return string.replace(replace, '\n').substr(1);
}

function isGenerator(obj) {
  return typeof obj.next === 'function' && typeof obj['throw'] === 'function';
}

function isGeneratorFunction(obj) {
  var constructor = obj.constructor;
  if (!constructor) return false;
  if (constructor.name === 'GeneratorFunction' || constructor.displayName === 'GeneratorFunction') return true;
  return isGenerator(constructor.prototype);
}